from __future__ import absolute_import
from .format import decide, FormatSelector
from rdflib.graph import Graph
from six import BytesIO
import itertools


class Decorator(object):
	def __init__(self, format_selector=None):
		self.format_selector = format_selector
		if self.format_selector is None:
			self.format_selector = FormatSelector()

	@staticmethod
	def _is_graph(obj):
		return isinstance(obj, Graph)

	@staticmethod
	def _get_graph(output):
		""" Given a WSGI response, check for a rdflib Graph """
		if Decorator._is_graph(output):	# single graph object
			return output

	def output(self, output, accepts, set_http_code, set_content_type):
		""" Formats a response from a WSGI app to handle any RDF graphs
		    If a view function returns a single RDF graph, serialize it based on Accept header
		    If it's not an RDF graph, return it without any special handling
		"""

		graph = Decorator._get_graph(output)
		if graph is not None:
			# decide the format
			output_mimetype, output_format = self.format_selector.decide(accepts, graph.context_aware)
			# requested content couldn't find anything
			if output_mimetype is None:
				set_http_code("406 Not Acceptable")
				return ['406 Not Acceptable'.encode('utf-8')]
			# explicitly mark text mimetypes as utf-8
			if 'text' in output_mimetype:
				output_mimetype = output_mimetype + '; charset=utf-8'

			# format the new response
			serialized = graph.serialize(format=output_format)
			set_content_type(output_mimetype)
			return [serialized]
		else:
			return output

	def decorate(self, app):
		""" Wraps a WSGI application to return formatted RDF graphs
		    Uses content negotiation to serialize the graph to the client-preferred format
		    Passes other content through unmodified
		"""
		from functools import wraps

		@wraps(app)
		def decorated(environ, start_response):
			# capture any start_response from the app
			app_response = {}
			app_response['status'] = "200 OK"
			app_response['headers'] = []
			app_response['written'] = BytesIO()
			def custom_start_response(status, headers, *args, **kwargs):
				app_response['status'] = status
				app_response['headers'] = headers
				app_response['args'] = args
				app_response['kwargs'] = kwargs
				return app_response['written'].write
			returned = app(environ, custom_start_response)

			# callbacks from the serialization
			def set_http_code(status):
				app_response['status'] = str(status)
			def set_header(header, value):
				app_response['headers'] = [(h,v) for (h,v) in app_response['headers'] if h.lower() != header.lower()]
				app_response['headers'].append((header, value))
			def set_content_type(content_type):
				set_header('Content-Type', content_type)

			# do the serialization
			accept = environ.get('HTTP_ACCEPT', '')
			new_return = self.output(returned, accept, set_http_code, set_content_type)

			# set the Vary header
			vary_headers = (v for (h,v) in app_response['headers'] if h.lower() == 'vary')
			vary_elements = list(itertools.chain(*[v.split(',') for v in vary_headers]))
			vary_elements = list(set([v.strip() for v in vary_elements]))
			if '*' not in vary_elements and 'accept' not in (v.lower() for v in vary_elements):
				vary_elements.append('Accept')
				set_header('Vary', ', '.join(vary_elements))

			# pass on the result to the parent WSGI server
			parent_writer = start_response(app_response['status'],
			                               app_response['headers'],
			                               *app_response.get('args', []),
			                               **app_response.get('kwargs', {}))
			written = app_response['written'].getvalue()
			if len(written) > 0:
				parent_writer(written)
			return new_return
		return decorated

	def __call__(self, app):
		""" Enables this class to be used as the decorator directly """
		return self.decorate(app)


_implicit_instance = Decorator()


def output(output, accepts, set_http_code, set_content_type):
	return _implicit_instance.output(output, accepts, set_http_code, set_content_type)
def returns_rdf(view):
	return _implicit_instance.decorate(view)



import mimeparse


DEFAULT_MIMETYPE = 'application/rdf+xml'	# default mimetype to return
WILDCARD = 'INVALID/MATCH'	# matches Accept:*/*
WILDCARD_MIMETYPE = 'application/rdf+xml'	# mimetype for wildcard


# What formats we support for serialization
formats = {
   'application/x-turtle': 'turtle',
   'text/turtle': 'turtle',
   'application/rdf+xml': 'xml',
   'application/trix': 'trix',
   'application/n-quads': 'nquads',
   'application/n-triples': 'nt',
   'text/n-triples': 'nt',
   'text/rdf+nt': 'nt',
   'application/n3': 'n3',
   'text/n3': 'n3',
   'text/rdf+n3': 'n3'
}
# the list of any mimetypes, unlocked if we have a context
all_mimetypes = list(formats.keys())
# the list of mimetypes that don't require a context
ctxless_mimetypes = [m for m in all_mimetypes if 'n-quads' not in m]


class FormatSelector(object):
	def __init__(self):
		# any extra formats that we support
		self.formats = {}
		# the list of any mimetypes, unlocked if we have a context
		self.all_mimetypes = []
		# the list of mimetypes that don't require a context
		self.ctxless_mimetypes = []
		# the default mimetype to use
		self.default_mimetype = None
		# the wildcard mimetype to use
		self.wildcard_mimetype = None

	def add_format(self, mimetype, format, requires_context=False):
		""" Registers a new format to be used in a graph's serialize call
		    If you've installed an rdflib serializer plugin, use this
		    to add it to the content negotiation system
		    Set requires_context=True if this format requires a context-aware graph
		"""
		self.formats[mimetype] = format
		if not requires_context:
			self.ctxless_mimetypes.append(mimetype)
		self.all_mimetypes.append(mimetype)

	def get_default_mimetype(self):
		""" Returns the default mimetype """
		mimetype = self.default_mimetype
		if mimetype is None:	# class inherits from module default
			mimetype = DEFAULT_MIMETYPE
		if mimetype is None:	# module is set to None?
			mimetype = 'application/rdf+xml'
		return mimetype

	def get_wildcard_mimetype(self):
		""" Returns the mimetype if the client sends */* """
		mimetype = self.wildcard_mimetype
		if mimetype is None:	# class inherits from module default
			mimetype = WILDCARD_MIMETYPE
		if mimetype is None:	# module is set to None?
			mimetype = 'application/rdf+xml'
		return mimetype

	def decide_mimetype(self, accepts, context_aware = False):
		""" Returns what mimetype the client wants to receive
		    Parses the given Accept header and returns the best one that
		    we know how to output
		    An empty Accept will default to application/rdf+xml
		    An Accept with */* use rdf+xml unless a better match is found
		    An Accept that doesn't match anything will return None
		"""
		mimetype = None
		# If the client didn't request a thing, use default
		if accepts is None or accepts.strip() == '':
			mimetype = self.get_default_mimetype()
			return mimetype

		# pick the mimetype
		if context_aware:
			mimetype = mimeparse.best_match(all_mimetypes + self.all_mimetypes + [WILDCARD], accepts)
		else:
			mimetype = mimeparse.best_match(ctxless_mimetypes + self.ctxless_mimetypes + [WILDCARD], accepts)
		if mimetype == '':
			mimetype = None

		# if browser sent */*
		if mimetype == WILDCARD:
			mimetype = self.get_wildcard_mimetype()

		return mimetype

	def get_serialize_format(self, mimetype):
		""" Get the serialization format for the given mimetype """
		format = self.formats.get(mimetype, None)
		if format is None:
			format = formats.get(mimetype, None)
		return format

	def decide(self, accepts, context_aware=False):
		""" Returns what (mimetype,format) the client wants to receive
		    Parses the given Accept header and picks the best one that
		    we know how to output
		    Returns (mimetype, format)
		    An empty Accept will default to rdf+xml
		    An Accept with */* use rdf+xml unless a better match is found
		    An Accept that doesn't match anything will return (None,None)
		    context_aware=True will allow nquad serialization
		"""
		mimetype = self.decide_mimetype(accepts, context_aware)
		# return what format to serialize as
		if mimetype is not None:
			return (mimetype, self.get_serialize_format(mimetype))
		else:
			# couldn't find a matching mimetype for the Accepts header
			return (None, None)

	def wants_rdf(self, accepts):
		""" Returns whether this client's Accept header indicates
		    that the client wants to receive RDF
		"""
		mimetype = mimeparse.best_match(all_mimetypes + self.all_mimetypes + [WILDCARD], accepts)
		return mimetype and mimetype != WILDCARD


_implicit_instance = FormatSelector()


def add_format(mimetype, format, requires_context=False):
	""" Registers a new format to be used in a graph's serialize call
	    If you've installed an rdflib serializer plugin, use this
	    to add it to the content negotiation system
	    Set requires_context=True if this format requires a context-aware graph
	"""
	global formats
	global ctxless_mimetypes
	global all_mimetypes
	formats[mimetype] = format
	if not requires_context:
		ctxless_mimetypes.append(mimetype)
	all_mimetypes.append(mimetype)

def decide(accepts, context_aware=False):
	return _implicit_instance.decide(accepts, context_aware)

def wants_rdf(accepts):
	""" Returns whether this client's Accept header indicates
	    that the client wants to receive RDF
	"""
	return _implicit_instance.wants_rdf(accepts)



from __future__ import absolute_import
from .format import decide, FormatSelector
from rdflib.graph import Graph


class ViewDecorator(object):
	def __init__(self, format_selector=None):
		self.format_selector = format_selector
		if self.format_selector is None:
			self.format_selector = FormatSelector()

	@classmethod
	def is_graph(cls, obj):
		""" Check whether this object is an rdflib Graph """
		return isinstance(obj, Graph)

	@classmethod
	def get_graph(cls, response):
		""" Given a view response, find the rdflib Graph, or None """
		if cls.is_graph(response):	# single graph object
			return response

	@classmethod
	def replace_graph(cls, response, serialized):
		""" Replace the rdflib Graph in a view response """
		if cls.is_graph(response):	# single graph object
			return serialized
		return response

	@classmethod
	def make_new_response(cls, old_response, mimetype, serialized):
		""" Return a new framework-specific response with the seralized data """
		raise NotImplementedError

	@classmethod
	def make_406_response(cls):
		""" Return the framework-specific HTTP 406 error """
		raise NotImplementedError

	@classmethod
	def get_accept(cls):
		""" Load the framework-specific Accept header """
		raise NotImplementedError

	def output(self, response, accepts):
		""" Formats a response from a view to handle any RDF graphs
		    If a view function returns an RDF graph, serialize it based on Accept header
		    If it's not an RDF graph, return it without any special handling
		"""
		graph = self.get_graph(response)
		if graph is not None:
			# decide the format
			mimetype, format = self.format_selector.decide(accepts, graph.context_aware)

			# requested content couldn't find anything
			if mimetype is None:
				return self.make_406_response()

			# explicitly mark text mimetypes as utf-8
			if 'text' in mimetype:
				mimetype = mimetype + '; charset=utf-8'

			# format the new response
			serialized = graph.serialize(format=format)
			response = self.make_new_response(response, mimetype, serialized)
			return response
		else:
			return response

	def decorate(self, view):
		""" Wraps a view function to return formatted RDF graphs
		    Uses content negotiation to serialize the graph to the client-preferred format
		    Passes other content through unmodified
		"""
		from functools import wraps

		@wraps(view)
		def decorated(*args, **kwargs):
			response = view(*args, **kwargs)
			accept = self.get_accept()
			return self.output(response, accept)
		return decorated

	def __call__(self, view):
		""" Enables this class to be used as the decorator directly """
		return self.decorate(view)



from __future__ import absolute_import
from .common_decorators import ViewDecorator
from rdflib.graph import Graph


class Decorator(ViewDecorator):
	@classmethod
	def get_graph(cls, response):
		""" Given a Flask response, find the rdflib Graph """
		if cls.is_graph(response):	# single graph object
			return response

		if hasattr(response, '__getitem__'):	# indexable tuple
			if len(response) > 0 and \
			   cls.is_graph(response[0]):	# graph object
				return response[0]

	@classmethod
	def replace_graph(cls, response, serialized):
		""" Replace the rdflib Graph in a Flask response """
		if cls.is_graph(response):	# single graph object
			return serialized

		if hasattr(response, '__getitem__'):	# indexable tuple
			if len(response) > 0 and \
			   cls.is_graph(response[0]):	# graph object
				return (serialized,) + response[1:]
		return response

	@classmethod
	def make_new_response(cls, old_response, mimetype, serialized):
		from flask import make_response
		final_output = cls.replace_graph(old_response, serialized)
		response = make_response(final_output)
		response.headers['Content-Type'] = mimetype
		response.headers['Vary'] = 'Accept'
		return response

	@classmethod
	def make_406_response(cls):
		return '406 Not Acceptable', 406

	@classmethod
	def get_accept(cls):
		from flask import request
		return request.headers.get('Accept', '')


_implicit_instance = Decorator()


def output(response, accepts):
	return _implicit_instance.output(response, accepts)
def returns_rdf(view):
	return _implicit_instance.decorate(view)



from . import bottle
from . import flask
from . import wsgi
from . import format
from .format import add_format, FormatSelector, wants_rdf
from .bottle import returns_rdf as bottle_rdf
from .flask import returns_rdf as flask_rdf
from .wsgi import returns_rdf as wsgi_rdf



from __future__ import absolute_import
from .common_decorators import ViewDecorator
from rdflib.graph import Graph


class Decorator(ViewDecorator):
	@classmethod
	def make_new_response(cls, old_response, mimetype, serialized):
		import bottle
		bottle.response.content_type = mimetype
		bottle.response.set_header('Vary', 'Accept')
		return serialized

	@classmethod
	def make_406_response(cls):
		import bottle
		bottle.abort(406, '406 Not Acceptable')

	@classmethod
	def get_accept(cls):
		import bottle
		return bottle.request.headers.get('Accept', '')


_implicit_instance = Decorator()


def output(output, accepts):
	return _implicit_instance.output(output, accepts)
def returns_rdf(view):
	return _implicit_instance.decorate(view)
